package org.codehaus.xma.mojo.packager
import org.apache.maven.plugin.Mojo;

import org.codehaus.groovy.maven.mojo.GroovyMojo
import org.codehaus.plexus.component.configurator.ComponentConfigurator;
import org.codehaus.plexus.component.configurator.expression.ExpressionEvaluator;
import org.codehaus.plexus.configuration.PlexusConfiguration;
import org.codehaus.plexus.configuration.xml.XmlPlexusConfiguration;
import org.codehaus.plexus.util.ReflectionUtils;
import org.codehaus.plexus.util.xml.Xpp3Dom;
import org.codehaus.plexus.util.xml.Xpp3DomBuilder;

import org.apache.tools.ant.types.FileSet;
import org.apache.tools.ant.DirectoryScanner;
import java.util.HashSet;


/**
 * Packager Mojo used to wrap and provide the existing xma ant facility as maven plugin.
 * <p>
 * The components to be packaged can be provided using four different methods:
 * <ul>
 * <li>Using the <xmp><components><xmp> element with a list of componentName=componentPackage style values
 *      (e.g. customeredit=org/openxma/dsl/reference/xma/customeredit/)
 * <li>Using the <xmp><components><xmp> element with a list of componentPackage values (e.g.
 *      org/openxma/dsl/reference/xma/customeredit/)
 * <li>Omitting the components element and defining an xmaXmlFilter element with an ANT fileset style filter value
 * <li>Omitting any of these, which results in packaging all XMA components in the source folder.
 * </ul>
 * <p>
 * Providing a filter as well as excplicit components results in packaging the components first,
 * then searching for further components using the filter and packageing those, too. The filter may
 * also include the explicitely defined components.
 * <p>
 * An example with all the possibilities used:
 * <p>
 * <xmp>
                    <xmaXmlFilter>**&amp;#047;*.xma</xmaXmlFilter>
                    <components>
                        <component>customeredit=org/openxma/dsl/reference/xma/customeredit/</component>
                        <component>org/openxma/dsl/reference/xma/customeredit2/</component>
                    </components>

 <xmp>
 * @goal package
 * @requiresDependencyResolution
 */
class PackageMojo extends GroovyMojo {
    /**
     * @parameter expression="${project}"
     * @required
     * @readonly
    */
    org.apache.maven.project.MavenProject project
   /**
     * @parameter expression="${session}"
     * @required
     * @readonly
    */
    org.apache.maven.execution.MavenSession session
    /**
     * @parameter expression="${basedir}/src/main/webapp/"
     */
    private String appDescriptorDirectory
    /**
     * @parameter
     */
    private String xmaXmlFilter
    /**
     * @parameter expression="xma-app.xml"
     */
    private String appDescriptorFileName
        /**
     * @parameter expression="plugins.xml"
     */
    private String pluginsDescriptorFileName
    /**
     * @parameter
     * @optional
     */
    private List plugins
    /**
     * @parameter
     * @optional
     */
    private List components
    /**
     * @parameter
     */
    private List clientLibraries
    /**
     * @parameter
     * @optional
     */
    private List clientIncludes
    /**
     * @parameter expression="${project.build.directory}/xma/components"
     */
    private String componentsBuildDirectory
    /**
     * @parameter expression="${project.build.directory}/xma/plugins"
     */
    private String pluginsBuildDirectory
    /**
     * @parameter expression="${project.build.directory}/xma/dependencies"
     */
    private String dependenciesBuildDirectory
    /**
     * @parameter expression="${project.build.directory}/xma/web"
     */
    private String xmaDirectory
    /**
     * @parameter expression="${project.build.directory}/xma/web/components"
     */
    private String componentsDirectory
    /**
     * @parameter expression="${project.build.directory}/xma/web/plugins"
     */
    private String pluginsDirectory
    /**
     * @parameter expression="${project.build.directory}/xma/web/clientrt"
     */
    private String clientRtDirectory

    /**
     * @parameter default-value="commonclient"
     */
    private String commonClientJarName

    void execute() {
        //removing all registered build listeners, including default (that writes to console)
        ant.project.getBuildListeners().each {
            ant.project.removeBuildListener(it)
        }
        // some preconditions..
        if (project.packaging!=/war/) {
            fail ("not a web project")
        }
        if (!appDescriptorFileExists() && !pluginsDescriptorFileExists()) {
            fail ("required appDescriptorFile '${appDescriptorFile.name}' not found in directory '${appDescriptorDirectory}'")
        }
        boolean fastDevelopmentMode = session.goals.any {it == "jetty:run" } || System.env["boot.fdm"]
        if (fastDevelopmentMode) {
              log.info  "openXma Fast-Development-Mode enabled"
        }
        createBuildDirectories()
        copyDependencyArtifacts()
        jarComponentsAndCreateChecksum(fastDevelopmentMode)
        jarClientIncludes()
        copyComponentsAndApplicationDescriptor(fastDevelopmentMode)
    }
    
    boolean appDescriptorFileExists() {
    	return new File(appDescriptorDirectory,appDescriptorFileName).exists()
    }
    
    boolean pluginsDescriptorFileExists() {
    	return new File(appDescriptorDirectory,pluginsDescriptorFileName).exists()
    }

    String createBuildDirectories() {
        log.info  "${project}"
        ant.mkdir(dir: dependenciesBuildDirectory)
		if (appDescriptorFileExists()) {
			ant.mkdir(dir: componentsBuildDirectory)
        	ant.mkdir(dir: "${componentsDirectory}")
        	ant.mkdir(dir: "${clientRtDirectory}")
		}
        if (pluginsDescriptorFileExists()) {
        	ant.mkdir(dir: pluginsBuildDirectory)
        	ant.mkdir(dir: "${pluginsDirectory}")
        }
        ant.taskdef(name: "xmachecksum", classname: "at.spardat.xma.boot.antext.XMAChecksum")
        ant.taskdef(name: "appdes",         classname: "at.spardat.xma.boot.antext.AppDescriptor")
    }

    void copyDependencyArtifacts() {
        log.info "Copy xma runtime dependencies.."
        project.dependencyArtifacts.plus(project.artifacts).unique().
            findAll({artifact -> !["test"].contains(artifact.scope) && !artifact.file.name.endsWith(".war") && !artifact.file.directory}).each({artifact ->
            def extensionStartIndex  = artifact.file.name.lastIndexOf('.')
            def fileName             = (extensionStartIndex>=0)?artifact.file.name.substring(0,extensionStartIndex):artifact.file.name
            def fileNameAndExtension = artifact.file.name.split('\\.') as List
            def fileNameExtension    = fileNameAndExtension.last()         
            if (fileName.split('-').size() <= 1) {
                log.warn("Artifact ${artifact.file.name} does not contain a valid '-' separator between artifactname and version!")
            }
            def fileNameWithoutVersion=null
            fileName.split('-').each({fileNamePart ->
                def indexOfFirstDot = fileNamePart.indexOf('.')
                def partBeforeFirstDot = (indexOfFirstDot>=0)?fileNamePart.substring(0,indexOfFirstDot):""                
                if (!partBeforeFirstDot.isNumber() && fileNamePart != "SNAPSHOT") {                    
                    fileNameWithoutVersion = null==fileNameWithoutVersion?fileNamePart:fileNameWithoutVersion+'-'+fileNamePart
            	}
            })             
            // log.info "Copy ${dependenciesBuildDirectory}/${fileNameWithoutVersion}.${fileNameExtension}"
            ant.copy(tofile: "${dependenciesBuildDirectory}/${fileNameWithoutVersion}.${fileNameExtension}",
                file:"${artifact.file}")
        })
    }

    void jarComponentsAndCreateChecksum(fastDevelopmentMode) {
        if (!fastDevelopmentMode) {
           	packComponents()
            packPlugins()
        }
        copyClientLibraries()
    }

	
	void packComponents() {
		if (appDescriptorFileExists()) {
	        log.info "Package component jars.."
			def processedComponents = new HashSet();
	        def componentNames = new HashSet();
	        components.each({ componentNameAndPackage ->
	            def componentSplitted = componentNameAndPackage.split('=')
	
	            def componentName
	            def componentPackage
	            if (componentSplitted.size() > 1) {
	                componentName = componentSplitted[0]
	                componentPackage = componentSplitted[1]
	            } else {
	                componentPackage = componentSplitted[0]
	                def componentPath = componentNameAndPackage.split('/')
	                componentName = componentPath[componentPath.length - 1]
	            }
	
	            if (componentPackage.endsWith('/')) {
	                componentPackage = componentPackage.substring(0, componentPackage.length() - 1 );
	            }
	
	            packComponent(componentName, componentPackage)
	            processedComponents.add(componentPackage);
	            if (componentNames.contains(componentName)) {
	                fail("Component name is not unique '${componentName}', the component must be renamed.")
	            }
	            componentNames.add(componentName);
	        })
	
	        if (components == null || components.isEmpty() || xmaXmlFilter != null) {
	            if (xmaXmlFilter == null) {
	                xmaXmlFilter = "**/*.xma";
	            }
	            log.info "Searching for xma components with: '${xmaXmlFilter}'"
	
	            FileSet fileSet = new FileSet();
	            fileSet.setProject(ant.project);
	            fileSet.setDir(new File( project.build.sourceDirectory ));
	            fileSet.setIncludes(xmaXmlFilter);
	
	            DirectoryScanner ds = fileSet.getDirectoryScanner(ant.project);
	            File dir = ds.getBasedir();
	
	            String[] filesInSet = ds.getIncludedFiles();
	            for (int i = 0; i < filesInSet.length; i++) {
	                def compXmaName = filesInSet[i]
	                compXmaName = compXmaName.replace('\\', '/')
	
	                log.info "Found xma component xml: '${compXmaName}'"
	
	                def componentPackage = compXmaName.substring(0,compXmaName.lastIndexOf('/'))
	                def componentName = compXmaName.substring(compXmaName.lastIndexOf('/')+1    ,compXmaName.lastIndexOf('.'))
	
	                if (processedComponents.contains(componentPackage)) {
	                    log.info "Component already packed, skipping now: '${componentPackage}'"
	                } else {
	                    packComponent(componentName, componentPackage);
	                    processedComponents.add(componentPackage);
	
	                    if (componentNames.contains(componentName)) {
	                        fail("Component name is not unique '${componentName}', the component must be renamed.")
	                    }
	                    componentNames.add(componentName);
	                }
	            }
	        }
	        if (componentNames.isEmpty()) {
	            fail("No component has been packed. This might not be what you want, so you maybe want to check your config.")
	        }
        }
	}
	
	void packComponent(componentName, componentPackage) {
        log.info "Package client/server component '${componentName}'.."
        ant.jar(
            destfile: "${componentsBuildDirectory}/${componentName.toLowerCase()}client.jar",
            basedir:  "${project.build.outputDirectory}",
            includes: "${componentPackage}/**",
            excludes: "${componentPackage}/**/server/*.*"
        )
        ant.xmachecksum(file: "${componentsBuildDirectory}/${componentName.toLowerCase()}client.jar")
        ant.jar(
            destfile: "${componentsBuildDirectory}/${componentName.toLowerCase()}server.jar",
            basedir:  "${project.build.outputDirectory}",
            includes: "${componentPackage}/**",
            excludes: "${componentPackage}/**/client/*.*"
        )
    }
	
	void packPlugins() {
        if (pluginsDescriptorFileExists()) {
        	log.info "Package plugin jars.."
	        def pluginNames = new HashSet();
	        plugins.each({ pluginNameAndPackage ->
	            def splittedName = pluginNameAndPackage.split('=')
	            def pluginName = splittedName[0]
	            def pluginPackages = splittedName[1]
	            if (!pluginNames.add(pluginName)) {
	                fail("Plugin name is not unique '${pluginName}', the plugin must be renamed.")
	            }
	            ant.jar(
		            destfile: "${pluginsBuildDirectory}/${pluginName.toLowerCase()}.jar",
		            basedir:  "${project.build.outputDirectory}",
		            includes: "${pluginPackages}"
		        )
	        	ant.xmachecksum(file: "${pluginsBuildDirectory}/${pluginName.toLowerCase()}.jar")
	        })
        }
	}
    
    void copyClientLibraries() {
		log.info "Copy client libraries.."
        clientLibraries.each({ jarNameAndDestinationDirectory ->
            def splitted = jarNameAndDestinationDirectory.split('=')
            def libraryName = splitted[0]
            def destinationFolder = splitted[1]
            def destFolderFile = new File(xmaDirectory, destinationFolder)
            if (!destFolderFile.exists()) {
                ant.mkdir(dir: xmaDirectory + File.separator + destinationFolder)
            }
            ant.copy(tofile: "${xmaDirectory}/${destinationFolder}/${libraryName}.jar", file:"${dependenciesBuildDirectory}/${libraryName}.jar")
            ant.xmachecksum(file: "${xmaDirectory}/${destinationFolder}/${libraryName}.jar")
        })
	}

    void jarClientIncludes() {
        if (clientIncludes!=null) {
	        log.info "Package common client includes.."
	        clientIncludes.each({ clientIncludePath -> log.info clientIncludePath })
	        ant.jar(destfile: "${componentsBuildDirectory}/${commonClientJarName}.jar", update: true) {
	                fileset(dir: "${project.build.outputDirectory}", includes: "${clientIncludes.join(',')}")
	        }
	        ant.xmachecksum(file: "${componentsBuildDirectory}/${commonClientJarName}.jar")
        }
    }

    void copyComponentsAndApplicationDescriptor(fastDevelopmentMode) {
        if (appDescriptorFileExists()) {
	    	log.info "Copy components jars to ${componentsDirectory}"
	        ant.copy(todir: "${componentsDirectory}", preservelastmodified:"true") {
	            fileset(dir: "${componentsBuildDirectory}", includes:"*.jar*", excludes:"*server.jar*")
	        }
	        log.info "Copy client jars to ${clientRtDirectory}"
	        if (new File("${dependenciesBuildDirectory}/xmabootrt.zip" ).exists()) {
	            ant.unzip(src:"${dependenciesBuildDirectory}/xmabootrt.zip" ,dest:"${dependenciesBuildDirectory}")
	        }
	        ant.copy(todir : "${clientRtDirectory}",    file : "${dependenciesBuildDirectory}/xmartclient.jar",preservelastmodified : "true"    )
	        ant.xmachecksum(file: "${clientRtDirectory}/xmartclient.jar")
	
	        ant.copy(todir : "${clientRtDirectory}",    file : "${dependenciesBuildDirectory}/epclient.jar",preservelastmodified : "true"    )
	        ant.xmachecksum(file: "${clientRtDirectory}/epclient.jar")	        	       
		}
		if (pluginsDescriptorFileExists()) {
        	log.info "Copy plugin jars to ${pluginsDirectory}"
	        ant.copy(todir: "${pluginsDirectory}", preservelastmodified:"true") {
	            fileset(dir: "${pluginsBuildDirectory}", includes:"*.jar*")
	        }
        }	
        log.info "Prepare xma application descriptors.."
        ant.copy(todir : "${xmaDirectory}", preservelastmodified : "true") {
            fileset(dir: "${appDescriptorDirectory}", includes:"**/*.*")
        }
        if (appDescriptorFileExists()) {
        	ant.appdes(dir: "${xmaDirectory}", file: "${appDescriptorFileName}", fdm: fastDevelopmentMode)
        }
        if (pluginsDescriptorFileExists()) {
        	ant.appdes(dir: "${xmaDirectory}", file: "${pluginsDescriptorFileName}", fdm: fastDevelopmentMode)
        }
    }
}


